#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# sys.path should contain absolute paths. With relative paths, the frozen application will fail to import modules when
# current working directory is changed.

import os
import sys
import tempfile

print('sys.stderr.encoding:', sys.stderr.encoding)
print('sys.path', sys.path)
print('CWD:', repr(os.getcwd()))

# Change working directory.
os.chdir(tempfile.gettempdir())
print('Changing working directory...')
print('CWD:', repr(os.getcwd()))

# Try import a module. It should fail
try:
    for pth in sys.path:
        if not os.path.isabs(pth):
            raise SystemExit('ERROR: sys.path not absolute')
    import datetime  # noqa: F401
except Exception:
    raise SystemExit('ERROR: sys.path not absolute')



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This requires a custom hand-crafted PyQt5 package, since some distributions do not include PyQt5.uic.port_v3
# for Python 2 and the other way round.
#
# PyQt5.uic.port_v2.test raises an AssertionError if imported under Python 3. But the hook should prohibit the
# inclusion of this module, so an ImportError should be raised. Therefore, the ImportError is what we expect
# and we just ignore it.

import PyQt5

# Ensure PyQt5 is our fake package.
assert PyQt5.__pyinstaller_fake_module_marker__ == '__pyinstaller_fake_module_marker__'

import PyQt5.uic.port_v3  # noqa: E402

print(PyQt5.uic.port_v3.__path__)

try:
    import PyQt5.uic.port_v2
    print(PyQt5.uic.port_v2.__path__)
except ImportError:
    print('PyQt5.uic.port_v2 not imported')
    pass



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re, string, traceback
from waflib import Logs, Utils, Errors


class PreprocError(Errors.WafError):
    pass


FILE_CACHE_SIZE = 100000
LINE_CACHE_SIZE = 100000
POPFILE = '-'
recursion_limit = 150
go_absolute = False
standard_includes = ['/usr/local/include', '/usr/include']
if Utils.is_win32:
    standard_includes = []
use_trigraphs = 0
strict_quotes = 0
g_optrans = {
    'not': '!',
    'not_eq': '!',
    'and': '&&',
    'and_eq': '&=',
    'or': '||',
    'or_eq': '|=',
    'xor': '^',
    'xor_eq': '^=',
    'bitand': '&',
    'bitor': '|',
    'compl': '~',
}
re_lines = re.compile(
    '^[ \t]*(?:#|%:)[ \t]*(ifdef|ifndef|if|else|elif|endif|include|import|define|undef|pragma)[ \t]*(.*)\r*$',
    re.IGNORECASE | re.MULTILINE
)
re_mac = re.compile(r"^[a-zA-Z_]\w*")
re_fun = re.compile('^[a-zA-Z_][a-zA-Z0-9_]*[(]')
re_pragma_once = re.compile(r'^\s*once\s*', re.IGNORECASE)
re_nl = re.compile('\\\\\r*\n', re.MULTILINE)
re_cpp = re.compile(r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"', re.DOTALL | re.MULTILINE)
trig_def = [('??' + a, b) for a, b in zip("=-/!'()<>", r'#~\|^[]{}')]
chr_esc = {'0': 0, 'a': 7, 'b': 8, 't': 9, 'n': 10, 'f': 11, 'v': 12, 'r': 13, '\\': 92, "'": 39}
NUM = 'i'
OP = 'O'
IDENT = 'T'
STR = 's'
CHAR = 'c'
tok_types = [NUM, STR, IDENT, OP]
exp_types = [
    r"""0[xX](?P<hex>[a-fA-F0-9]+)(?P<qual1>[uUlL]*)|L*?'(?P<char>(\\.|[^\\'])+)'|(?P<n1>\d+)[Ee](?P<exp0>[+-]*?\d+)(?P<float0>[fFlL]*)|(?P<n2>\d*\.\d+)([Ee](?P<exp1>[+-]*?\d+))?(?P<float1>[fFlL]*)|(?P<n4>\d+\.\d*)([Ee](?P<exp2>[+-]*?\d+))?(?P<float2>[fFlL]*)|(?P<oct>0*)(?P<n0>\d+)(?P<qual2>[uUlL]*)""",
    r'L?"([^"\\]|\\.)*"',
    r'[a-zA-Z_]\w*',
    r'%:%:|<<=|>>=|\.\.\.|<<|<%|<:|<=|>>|>=|\+\+|\+=|--|->|-=|\*=|/=|%:|%=|%>|==|&&|&=|\|\||\|=|\^=|:>|!=|##|[\(\)\{\}\[\]<>\?\|\^\*\+&=:!#;,%/\-\?\~\.]',
]
re_clexer = re.compile('|'.join(["(?P<%s>%s)" % (name, part) for name, part in zip(tok_types, exp_types)]), re.M)
accepted = 'a'
ignored = 'i'
undefined = 'u'
skipped = 's'


def repl(m):
    s = m.group()
    if s[0] == '/':
        return ' '
    return s


prec = {}
ops = ['* / %', '+ -', '<< >>', '< <= >= >', '== !=', '& | ^', '&& ||', ',']
for x, syms in enumerate(ops):
    for u in syms.split():
        prec[u] = x


def reduce_nums(val_1, val_2, val_op):
    try:
        a = 0 + val_1
    except TypeError:
        a = int(val_1)
    try:
        b = 0 + val_2
    except TypeError:
        b = int(val_2)
    d = val_op
    if d == '%':
        c = a % b
    elif d == '+':
        c = a + b
    elif d == '-':
        c = a - b
    elif d == '*':
        c = a * b
    elif d == '/':
        c = a / b
    elif d == '^':
        c = a ^ b
    elif d == '==':
        c = int(a == b)
    elif d == '|' or d == 'bitor':
        c = a | b
    elif d == '||' or d == 'or':
        c = int(a or b)
    elif d == '&' or d == 'bitand':
        c = a & b
    elif d == '&&' or d == 'and':
        c = int(a and b)
    elif d == '!=' or d == 'not_eq':
        c = int(a != b)
    elif d == '^' or d == 'xor':
        c = int(a ^ b)
    elif d == '<=':
        c = int(a <= b)
    elif d == '<':
        c = int(a < b)
    elif d == '>':
        c = int(a > b)
    elif d == '>=':
        c = int(a >= b)
    elif d == '<<':
        c = a << b
    elif d == '>>':
        c = a >> b
    else:
        c = 0
    return c


def get_num(lst):
    if not lst:
        raise PreprocError('empty list for get_num')
    (p, v) = lst[0]
    if p == OP:
        if v == '(':
            count_par = 1
            i = 1
            while i < len(lst):
                (p, v) = lst[i]
                if p == OP:
                    if v == ')':
                        count_par -= 1
                        if count_par == 0:
                            break
                    elif v == '(':
                        count_par += 1
                i += 1
            else:
                raise PreprocError('rparen expected %r' % lst)
            (num, _) = get_term(lst[1:i])
            return (num, lst[i + 1:])
        elif v == '+':
            return get_num(lst[1:])
        elif v == '-':
            num, lst = get_num(lst[1:])
            return (reduce_nums('-1', num, '*'), lst)
        elif v == '!':
            num, lst = get_num(lst[1:])
            return (int(not int(num)), lst)
        elif v == '~':
            num, lst = get_num(lst[1:])
            return (~int(num), lst)
        else:
            raise PreprocError('Invalid op token %r for get_num' % lst)
    elif p == NUM:
        return v, lst[1:]
    elif p == IDENT:
        return 0, lst[1:]
    else:
        raise PreprocError('Invalid token %r for get_num' % lst)


def get_term(lst):
    if not lst:
        raise PreprocError('empty list for get_term')
    num, lst = get_num(lst)
    if not lst:
        return (num, [])
    (p, v) = lst[0]
    if p == OP:
        if v == ',':
            return get_term(lst[1:])
        elif v == '?':
            count_par = 0
            i = 1
            while i < len(lst):
                (p, v) = lst[i]
                if p == OP:
                    if v == ')':
                        count_par -= 1
                    elif v == '(':
                        count_par += 1
                    elif v == ':':
                        if count_par == 0:
                            break
                i += 1
            else:
                raise PreprocError('rparen expected %r' % lst)
            if int(num):
                return get_term(lst[1:i])
            else:
                return get_term(lst[i + 1:])
        else:
            num2, lst = get_num(lst[1:])
            if not lst:
                num2 = reduce_nums(num, num2, v)
                return get_term([(NUM, num2)] + lst)
            p2, v2 = lst[0]
            if p2 != OP:
                raise PreprocError('op expected %r' % lst)
            if prec[v2] >= prec[v]:
                num2 = reduce_nums(num, num2, v)
                return get_term([(NUM, num2)] + lst)
            else:
                num3, lst = get_num(lst[1:])
                num3 = reduce_nums(num2, num3, v2)
                return get_term([(NUM, num), (p, v), (NUM, num3)] + lst)
    raise PreprocError('cannot reduce %r' % lst)


def reduce_eval(lst):
    num, lst = get_term(lst)
    return (NUM, num)


def stringize(lst):
    lst = [str(v2) for (p2, v2) in lst]
    return "".join(lst)


def paste_tokens(t1, t2):
    p1 = None
    if t1[0] == OP and t2[0] == OP:
        p1 = OP
    elif t1[0] == IDENT and (t2[0] == IDENT or t2[0] == NUM):
        p1 = IDENT
    elif t1[0] == NUM and t2[0] == NUM:
        p1 = NUM
    if not p1:
        raise PreprocError('tokens do not make a valid paste %r and %r' % (t1, t2))
    return (p1, t1[1] + t2[1])


def reduce_tokens(lst, defs, ban=[]):
    i = 0
    while i < len(lst):
        (p, v) = lst[i]
        if p == IDENT and v == "defined":
            del lst[i]
            if i < len(lst):
                (p2, v2) = lst[i]
                if p2 == IDENT:
                    if v2 in defs:
                        lst[i] = (NUM, 1)
                    else:
                        lst[i] = (NUM, 0)
                elif p2 == OP and v2 == '(':
                    del lst[i]
                    (p2, v2) = lst[i]
                    del lst[i]
                    if v2 in defs:
                        lst[i] = (NUM, 1)
                    else:
                        lst[i] = (NUM, 0)
                else:
                    raise PreprocError('Invalid define expression %r' % lst)
        elif p == IDENT and v in defs:
            if isinstance(defs[v], str):
                a, b = extract_macro(defs[v])
                defs[v] = b
            macro_def = defs[v]
            to_add = macro_def[1]
            if isinstance(macro_def[0], list):
                del lst[i]
                accu = to_add[:]
                reduce_tokens(accu, defs, ban + [v])
                for tmp in accu:
                    lst.insert(i, tmp)
                    i += 1
            else:
                args = []
                del lst[i]
                if i >= len(lst):
                    raise PreprocError('expected ( after %r (got nothing)' % v)
                (p2, v2) = lst[i]
                if p2 != OP or v2 != '(':
                    raise PreprocError('expected ( after %r' % v)
                del lst[i]
                one_param = []
                count_paren = 0
                while i < len(lst):
                    p2, v2 = lst[i]
                    del lst[i]
                    if p2 == OP and count_paren == 0:
                        if v2 == '(':
                            one_param.append((p2, v2))
                            count_paren += 1
                        elif v2 == ')':
                            if one_param:
                                args.append(one_param)
                            break
                        elif v2 == ',':
                            if not one_param:
                                raise PreprocError('empty param in funcall %r' % v)
                            args.append(one_param)
                            one_param = []
                        else:
                            one_param.append((p2, v2))
                    else:
                        one_param.append((p2, v2))
                        if v2 == '(':
                            count_paren += 1
                        elif v2 == ')':
                            count_paren -= 1
                else:
                    raise PreprocError('malformed macro')
                accu = []
                arg_table = macro_def[0]
                j = 0
                while j < len(to_add):
                    (p2, v2) = to_add[j]
                    if p2 == OP and v2 == '#':
                        if j + 1 < len(to_add) and to_add[j + 1][0] == IDENT and to_add[j + 1][1] in arg_table:
                            toks = args[arg_table[to_add[j + 1][1]]]
                            accu.append((STR, stringize(toks)))
                            j += 1
                        else:
                            accu.append((p2, v2))
                    elif p2 == OP and v2 == '##':
                        if accu and j + 1 < len(to_add):
                            t1 = accu[-1]
                            if to_add[j + 1][0] == IDENT and to_add[j + 1][1] in arg_table:
                                toks = args[arg_table[to_add[j + 1][1]]]
                                if toks:
                                    accu[-1] = paste_tokens(t1, toks[0])
                                    accu.extend(toks[1:])
                                else:
                                    accu.append((p2, v2))
                                    accu.extend(toks)
                            elif to_add[j + 1][0] == IDENT and to_add[j + 1][1] == '__VA_ARGS__':
                                va_toks = []
                                st = len(macro_def[0])
                                pt = len(args)
                                for x in args[pt - st + 1:]:
                                    va_toks.extend(x)
                                    va_toks.append((OP, ','))
                                if va_toks:
                                    va_toks.pop()
                                if len(accu) > 1:
                                    (p3, v3) = accu[-1]
                                    (p4, v4) = accu[-2]
                                    if v3 == '##':
                                        accu.pop()
                                        if v4 == ',' and pt < st:
                                            accu.pop()
                                accu += va_toks
                            else:
                                accu[-1] = paste_tokens(t1, to_add[j + 1])
                            j += 1
                        else:
                            accu.append((p2, v2))
                    elif p2 == IDENT and v2 in arg_table:
                        toks = args[arg_table[v2]]
                        reduce_tokens(toks, defs, ban + [v])
                        accu.extend(toks)
                    else:
                        accu.append((p2, v2))
                    j += 1
                reduce_tokens(accu, defs, ban + [v])
                for x in range(len(accu) - 1, -1, -1):
                    lst.insert(i, accu[x])
        i += 1


def eval_macro(lst, defs):
    reduce_tokens(lst, defs, [])
    if not lst:
        raise PreprocError('missing tokens to evaluate')
    if lst:
        p, v = lst[0]
        if p == IDENT and v not in defs:
            raise PreprocError('missing macro %r' % lst)
    p, v = reduce_eval(lst)
    return int(v) != 0


def extract_macro(txt):
    t = tokenize(txt)
    if re_fun.search(txt):
        p, name = t[0]
        p, v = t[1]
        if p != OP:
            raise PreprocError('expected (')
        i = 1
        pindex = 0
        params = {}
        prev = '('
        while 1:
            i += 1
            p, v = t[i]
            if prev == '(':
                if p == IDENT:
                    params[v] = pindex
                    pindex += 1
                    prev = p
                elif p == OP and v == ')':
                    break
                else:
                    raise PreprocError('unexpected token (3)')
            elif prev == IDENT:
                if p == OP and v == ',':
                    prev = v
                elif p == OP and v == ')':
                    break
                else:
                    raise PreprocError('comma or ... expected')
            elif prev == ',':
                if p == IDENT:
                    params[v] = pindex
                    pindex += 1
                    prev = p
                elif p == OP and v == '...':
                    raise PreprocError('not implemented (1)')
                else:
                    raise PreprocError('comma or ... expected (2)')
            elif prev == '...':
                raise PreprocError('not implemented (2)')
            else:
                raise PreprocError('unexpected else')
        return (name, [params, t[i + 1:]])
    else:
        (p, v) = t[0]
        if len(t) > 1:
            return (v, [[], t[1:]])
        else:
            return (v, [[], [('T', '')]])


re_include = re.compile(r'^\s*(<(?:.*)>|"(?:.*)")')


def extract_include(txt, defs):
    m = re_include.search(txt)
    if m:
        txt = m.group(1)
        return txt[0], txt[1:-1]
    toks = tokenize(txt)
    reduce_tokens(toks, defs, ['waf_include'])
    if not toks:
        raise PreprocError('could not parse include %r' % txt)
    if len(toks) == 1:
        if toks[0][0] == STR:
            return '"', toks[0][1]
    else:
        if toks[0][1] == '<' and toks[-1][1] == '>':
            ret = '<', stringize(toks).lstrip('<').rstrip('>')
            return ret
    raise PreprocError('could not parse include %r' % txt)


def parse_char(txt):
    if not txt:
        raise PreprocError('attempted to parse a null char')
    if txt[0] != '\\':
        return ord(txt)
    c = txt[1]
    if c == 'x':
        if len(txt) == 4 and txt[3] in string.hexdigits:
            return int(txt[2:], 16)
        return int(txt[2:], 16)
    elif c.isdigit():
        if c == '0' and len(txt) == 2:
            return 0
        for i in 3, 2, 1:
            if len(txt) > i and txt[1:1 + i].isdigit():
                return (1 + i, int(txt[1:1 + i], 8))
    else:
        try:
            return chr_esc[c]
        except KeyError:
            raise PreprocError('could not parse char literal %r' % txt)


def tokenize(s):
    return tokenize_private(s)[:]


def tokenize_private(s):
    ret = []
    for match in re_clexer.finditer(s):
        m = match.group
        for name in tok_types:
            v = m(name)
            if v:
                if name == IDENT:
                    if v in g_optrans:
                        name = OP
                    elif v.lower() == "true":
                        v = 1
                        name = NUM
                    elif v.lower() == "false":
                        v = 0
                        name = NUM
                elif name == NUM:
                    if m('oct'):
                        v = int(v, 8)
                    elif m('hex'):
                        v = int(m('hex'), 16)
                    elif m('n0'):
                        v = m('n0')
                    else:
                        v = m('char')
                        if v:
                            v = parse_char(v)
                        else:
                            v = m('n2') or m('n4')
                elif name == OP:
                    if v == '%:':
                        v = '#'
                    elif v == '%:%:':
                        v = '##'
                elif name == STR:
                    v = v[1:-1]
                ret.append((name, v))
                break
    return ret


def format_defines(lst):
    ret = []
    for y in lst:
        if y:
            pos = y.find('=')
            if pos == -1:
                ret.append(y)
            elif pos > 0:
                ret.append('%s %s' % (y[:pos], y[pos + 1:]))
            else:
                raise ValueError('Invalid define expression %r' % y)
    return ret


class c_parser(object):
    def __init__(self, nodepaths=None, defines=None):
        self.lines = []
        if defines is None:
            self.defs = {}
        else:
            self.defs = dict(defines)
        self.state = []
        self.count_files = 0
        self.currentnode_stack = []
        self.nodepaths = nodepaths or []
        self.nodes = []
        self.names = []
        self.curfile = ''
        self.ban_includes = set()
        self.listed = set()

    def cached_find_resource(self, node, filename):
        try:
            cache = node.ctx.preproc_cache_node
        except AttributeError:
            cache = node.ctx.preproc_cache_node = Utils.lru_cache(FILE_CACHE_SIZE)
        key = (node, filename)
        try:
            return cache[key]
        except KeyError:
            ret = node.find_resource(filename)
            if ret:
                if getattr(ret, 'children', None):
                    ret = None
                elif ret.is_child_of(node.ctx.bldnode):
                    tmp = node.ctx.srcnode.search_node(ret.path_from(node.ctx.bldnode))
                    if tmp and getattr(tmp, 'children', None):
                        ret = None
            cache[key] = ret
            return ret

    def tryfind(self, filename, kind='"', env=None):
        if filename.endswith('.moc'):
            self.names.append(filename)
            return None
        self.curfile = filename
        found = None
        if kind == '"':
            if env.MSVC_VERSION:
                for n in reversed(self.currentnode_stack):
                    found = self.cached_find_resource(n, filename)
                    if found:
                        break
            else:
                found = self.cached_find_resource(self.currentnode_stack[-1], filename)
        if not found:
            for n in self.nodepaths:
                found = self.cached_find_resource(n, filename)
                if found:
                    break
        listed = self.listed
        if found and not found in self.ban_includes:
            if found not in listed:
                listed.add(found)
                self.nodes.append(found)
            self.addlines(found)
        else:
            if filename not in listed:
                listed.add(filename)
                self.names.append(filename)
        return found

    def filter_comments(self, node):
        code = node.read()
        if use_trigraphs:
            for (a, b) in trig_def:
                code = code.split(a).join(b)
        code = re_nl.sub('', code)
        code = re_cpp.sub(repl, code)
        return re_lines.findall(code)

    def parse_lines(self, node):
        try:
            cache = node.ctx.preproc_cache_lines
        except AttributeError:
            cache = node.ctx.preproc_cache_lines = Utils.lru_cache(LINE_CACHE_SIZE)
        try:
            return cache[node]
        except KeyError:
            cache[node] = lines = self.filter_comments(node)
            lines.append((POPFILE, ''))
            lines.reverse()
            return lines

    def addlines(self, node):
        self.currentnode_stack.append(node.parent)
        self.count_files += 1
        if self.count_files > recursion_limit:
            raise PreprocError('recursion limit exceeded')
        if Logs.verbose:
            Logs.debug('preproc: reading file %r', node)
        try:
            lines = self.parse_lines(node)
        except EnvironmentError:
            raise PreprocError('could not read the file %r' % node)
        except Exception:
            if Logs.verbose > 0:
                Logs.error('parsing %r failed %s', node, traceback.format_exc())
        else:
            self.lines.extend(lines)

    def start(self, node, env):
        Logs.debug('preproc: scanning %s (in %s)', node.name, node.parent.name)
        self.current_file = node
        self.addlines(node)
        if env.DEFINES:
            lst = format_defines(env.DEFINES)
            lst.reverse()
            self.lines.extend([('define', x) for x in lst])
        while self.lines:
            (token, line) = self.lines.pop()
            if token == POPFILE:
                self.count_files -= 1
                self.currentnode_stack.pop()
                continue
            try:
                state = self.state
                if token[:2] == 'if':
                    state.append(undefined)
                elif token == 'endif':
                    state.pop()
                if token[0] != 'e':
                    if skipped in self.state or ignored in self.state:
                        continue
                if token == 'if':
                    ret = eval_macro(tokenize(line), self.defs)
                    if ret:
                        state[-1] = accepted
                    else:
                        state[-1] = ignored
                elif token == 'ifdef':
                    m = re_mac.match(line)
                    if m and m.group() in self.defs:
                        state[-1] = accepted
                    else:
                        state[-1] = ignored
                elif token == 'ifndef':
                    m = re_mac.match(line)
                    if m and m.group() in self.defs:
                        state[-1] = ignored
                    else:
                        state[-1] = accepted
                elif token == 'include' or token == 'import':
                    (kind, inc) = extract_include(line, self.defs)
                    self.current_file = self.tryfind(inc, kind, env)
                    if token == 'import':
                        self.ban_includes.add(self.current_file)
                elif token == 'elif':
                    if state[-1] == accepted:
                        state[-1] = skipped
                    elif state[-1] == ignored:
                        if eval_macro(tokenize(line), self.defs):
                            state[-1] = accepted
                elif token == 'else':
                    if state[-1] == accepted:
                        state[-1] = skipped
                    elif state[-1] == ignored:
                        state[-1] = accepted
                elif token == 'define':
                    try:
                        self.defs[self.define_name(line)] = line
                    except AttributeError:
                        raise PreprocError('Invalid define line %r' % line)
                elif token == 'undef':
                    m = re_mac.match(line)
                    if m and m.group() in self.defs:
                        self.defs.__delitem__(m.group())
                elif token == 'pragma':
                    if re_pragma_once.match(line.lower()):
                        self.ban_includes.add(self.current_file)
            except Exception as e:
                if Logs.verbose:
                    Logs.debug('preproc: line parsing failed (%s): %s %s', e, line, traceback.format_exc())

    def define_name(self, line):
        return re_mac.match(line).group()


def scan(task):
    try:
        incn = task.generator.includes_nodes
    except AttributeError:
        raise Errors.WafError('%r is missing a feature such as "c", "cxx" or "includes": ' % task.generator)
    if go_absolute:
        nodepaths = incn + [task.generator.bld.root.find_dir(x) for x in standard_includes]
    else:
        nodepaths = [x for x in incn if x.is_child_of(x.ctx.srcnode) or x.is_child_of(x.ctx.bldnode)]
    tmp = c_parser(nodepaths)
    tmp.start(task.inputs[0], task.env)
    return (tmp.nodes, tmp.names)



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Tests for PyInstaller.depend.bytecode
"""

import re
from types import CodeType
from textwrap import dedent, indent
import operator

from PyInstaller import compat
from PyInstaller.depend.bytecode import (
    function_calls,
    recursive_function_calls,
    any_alias,
    finditer,
    _cleanup_bytecode_string,  # used for sanity check in test_finditer()
)


def compile_(x):
    return compile(dedent(x), "<no file>", "exec")


def many_int_constants():
    """
    Generate Python code that includes >256 integer constants.
    """
    # NOTE: in python >= 3.14.0a2, integer arguments smaller than 256 are pushed directly to stack, without using a
    # co_consts. Therefore, to effectively use >256 constants, we need to generate >512 integer arguments.
    return "".join(f'a = {i}\n' for i in range(600))


def many_str_constants():
    """
    Generate Python code that includes >256 string constants.
    """
    return "".join(f'a = "val_{i}"\n' for i in range(300))


def many_globals():
    """
    Generate Python code that includes >256 of global identifiers.
    """
    return " = ".join(f"a_{i}" for i in range(300)) + " = 'hello'\n"


def many_arguments():
    """
    Generate a function call taking >256 arguments.
    """
    return 'foo({})\n'.format(", ".join(map(str, range(300))))


def in_a_function(body):
    """
    Define a function called function() containing **body**.
    """
    return "def function():\n" + indent(body, "    ") + "\n"


# Sanity check that no fancy bytecode optimisation causes code from either of the above functions to be automatically
# removed as redundant by the compiler.


def test_many_int_constants():
    code: CodeType = compile_(many_int_constants())
    # Only the variable name 'a'.
    assert code.co_names == ('a',)

    # In python >= 3.14.0a2, LOAD_SMALL_INT instruction is used to push integers smaller than 256 on the stack, and
    # co_consts is used for larger values (in combination with LOAD_CONST / LOAD_CONST_IMMORTAL).
    # In earlier python versions, co_consts is used for all constants.
    if compat.is_py314:
        # In 3.14.0a7 the behavior was changed (by 55815a6); it seems that the value of very first LOAD_SMALL_INT is
        # added to co_consts for some reason. This does not happen if there is preceding LOAD_CONST (e.g., if there
        # is a docstring present before the code). In case this behavior change was unintended or is changed further,
        # check if co_consts contains 0 at first index and adjust expected length accordingly...
        expected_length = 601 - 256  # (600 - 256) integers plus a `None` return.
        if code.co_consts[0] == 0:
            expected_length += 1
        assert len(code.co_consts) == expected_length
    else:
        # 600 integers plus a 'None' return.
        assert len(code.co_consts) == 601


def test_many_str_constants():
    code: CodeType = compile_(many_str_constants())
    # Only the variable name 'a'.
    assert code.co_names == ('a',)

    # 300 string constants plus a 'None' return.
    assert len(code.co_consts) == 301


def test_many_globals():
    code: CodeType = compile_(many_globals())
    assert len(code.co_names) == 300
    assert len(code.co_consts) == 2


def test_global_functions():
    """
    Test finding function calls in the global namespace.
    """

    # The simplest possible function call.
    code = compile_("foo()")
    assert function_calls(code) == [('foo', [])]

    # With arguments.
    code = compile_("foo('a')")
    assert function_calls(code) == [('foo', ['a'])]

    # Having >256 constants will take us into extended arg territory where multiple byte-pair instructions are needed
    # to reference the constant. If everything works, we should not notice the difference.
    code = compile_(many_int_constants() + "foo(.123)")
    assert function_calls(code) == [('foo', [.123])]

    code = compile_(many_str_constants() + "foo(.321)")
    assert function_calls(code) == [('foo', [.321])]

    # Similarly, >256 global names also requires special handling.
    code = compile_(many_globals() + "foo(.456)")
    assert function_calls(code) == [('foo', [.456])]

    # And the unlikely case of >256 arguments to one function call.
    #
    # NOTE: with python >= 3.14.0a5, this creates a list with a sequence
    # of BUILD_LIST, LOAD*, LIST_APPEND opcodes, followed by a
    # CALL_INTRINSIC_1 opcode with INTRINSIC_LIST_TO_TUPLE argument,
    # and CALL_FUNCTION_EX opcode.
    #
    # Since we have no real use case for such lists, perform the
    # test only on earlier python versions.
    if not compat.is_py314:
        code = compile_(many_arguments())
        assert function_calls(code) == [('foo', list(range(300)))]

    # For loops, if statements should work. The iterable in a comprehension loop works but the statement to be executed
    # repeatedly gets its own code object and therefore requires recursion (tested later).
    code = compile_(
        """
        for i in foo(1, 2):
            a = bar(3)
            if wop(4) > whip(5):
                whiz(6)
                [7 for i in whallop(8)]
        """
    )
    assert function_calls(code) == [
        ("foo", [1, 2]),
        ("bar", [3]),
        ("wop", [4]),
        ("whip", [5]),
        ("whiz", [6]),
        ("whallop", [8]),
    ]

    # These are not supported but should be silently ignored without unintentional errors:
    assert function_calls(compile_("foo(x)")) == []
    assert function_calls(compile_("foo(a='3')")) == []
    assert function_calls(compile_("foo(bar())")) == [('bar', [])]

    # Python's compiler evaluates arithmetic.
    out = function_calls(compile_("foo(1 + 1)"))
    if out:
        # However, I will not bank on this being guaranteed behaviour.
        assert out == [("foo", [2])]

    assert function_calls(compile_("foo.bar()")) == [("foo.bar", [])]
    assert function_calls(compile_("foo.bar.pop.whack('a', 'b')")) == [("foo.bar.pop.whack", ['a', 'b'])]


def test_nested_codes():
    """
    Test function_calls() on global functions in nested code objects (bodies of other functions).
    """

    # The following compile() creates 3 code objects:
    #   - A global code.
    #   = The contents of foo().
    #   - And the body of the embedded lambda.

    code = compile_(
        """
        def foo():
            bar()
            whoop = lambda : fizz(3)
            return range(10)
        """
    )
    # There are no function calls in the global code.
    assert function_calls(code) == []

    # Get the body of foo().
    foo_code, = (i for i in code.co_consts if isinstance(i, CodeType))
    # foo() contains bar() and the iterable of the comprehension loop.
    assert function_calls(foo_code) == [('bar', []), ('range', [10])]

    # Get the body of the embedded lambda.
    lambda_code = next(i for i in foo_code.co_consts if isinstance(i, CodeType))
    # This contains fizz(3).
    assert function_calls(lambda_code) == [('fizz', [3])]

    assert recursive_function_calls(code) == {
        code: [],
        foo_code: [('bar', []), ('range', [10])],
        lambda_code: [('fizz', [3])],
    }


def test_local_functions():
    """
    Test on purely local functions. I.e., the function was imported and called inside the body of another function.
    """
    code_ = compile_(
        in_a_function(
            """
            a = 3
            import foo, zap
            zap.pop(), foo.bar()
            """
        )
    )

    code: CodeType
    code, = (i for i in code_.co_consts if isinstance(i, CodeType))

    # This test may mistakenly pass if co_names and co_varnames can be mixed up.
    # Ensure co_names[i] != co_varnames[i] holds for all `i`.
    assert all(map(operator.ne, code.co_names, code.co_varnames))

    assert function_calls(code) == [('zap.pop', []), ('foo.bar', [])]


def test_any_alias():
    assert tuple(any_alias("foo.bar.pop")) == ("foo.bar.pop", "bar.pop", "pop")


def test_finditer():
    """
    Test that bytecode.finditer() yields matches only that start on an even byte (``match.start() % 2 == 0``).

    There are 3 permutations here when considering a match:
    - A match starts on an even byte:
        That's good! Include that sequence.
    - A single character match starts on an odd byte:
        Ignore it. It's a false positive.
    - A multi-character match starts on an odd byte:
        This match will be a false positive but there may be a genuine match shortly afterwards (in the case of the
        # test below - it'll be the next character) which overlaps with this one so we must override regex's
        behaviour of ignoring overlapping matches to prevent these from getting lost.
    """

    # separator: 0xFF
    sample_string = b"0123\xFF4567\xFF890\xFF12\xFF3\xFF4"

    # Sanity check - ensure that none of the characters in the sample string coincide with the opcodes that `finditer()`
    # filters out via call to `_cleanup_bytecode_string()` (e.g., CACHE, PUSH_NULL), If that is the case, we need to
    # pick up new separator to avoid disturbing the test. For example, original separator was space character, but its
    # ordinal code (32) coincides with PUSH_NULL opcode in python 3.14.0a7.
    assert sample_string == _cleanup_bytecode_string(sample_string), \
        "One of characters in input string coincides with filtered-out opcode!"

    matches = list(finditer(re.compile(rb"\d+"), sample_string))
    aligned = [i.group() for i in matches]
    assert aligned == [b"0123", b"567", b"890", b"12"]



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks import collect_submodules, can_import_module
from PyInstaller.utils.hooks.setuptools import setuptools_info

hiddenimports = []
excludedimports = ['__main__']

# pkg_resources keeps vendored modules in its _vendor subpackage, and does sys.meta_path based import magic to expose
# them as pkg_resources.extern.*
#
# With setuptools >= 71.0, pkg_resources ceased to vendor packages, because vendoring is now done at the setuptools
# level.
if setuptools_info.available and setuptools_info.version < (71, 0, 0):
    # The `railroad` package is an optional requirement for `pyparsing`. `pyparsing.diagrams` depends on `railroad`, so
    # filter it out when `railroad` is not available.
    if can_import_module('railroad'):
        hiddenimports += collect_submodules('pkg_resources._vendor')
    else:
        hiddenimports += collect_submodules(
            'pkg_resources._vendor', filter=lambda name: 'pkg_resources._vendor.pyparsing.diagram' not in name
        )

    # pkg_resources v45.0 dropped support for Python 2 and added this module printing a warning. We could save some
    # bytes if we would replace this by a fake module.
    if setuptools_info.version >= (45, 0, 0) and setuptools_info.version < (49, 1, 1):
        hiddenimports += ['pkg_resources.py2_warn']

    # As of v60.7, setuptools vendored jaraco and has pkg_resources use it. Currently, the pkg_resources._vendor.jaraco
    # namespace package cannot be automatically scanned due to limited support for pure namespace packages in our hook
    # utilities.
    #
    # In setuptools 60.7.0, the vendored jaraco.text package included "Lorem Ipsum.txt" data file, which also has to be
    # collected. However, the presence of the data file (and the resulting directory hierarchy) confuses the importer's
    # redirection logic; instead of trying to work-around that, tell user to upgrade or downgrade their setuptools.
    if setuptools_info.version == (60, 7, 0):
        raise SystemExit(
            "ERROR: Setuptools 60.7.0 is incompatible with PyInstaller. "
            "Downgrade to an earlier version or upgrade to a later version."
        )
    # In setuptools 60.7.1, the "Lorem Ipsum.txt" data file was dropped from the vendored jaraco.text package, so we can
    # accommodate it with couple of hidden imports.
    elif setuptools_info.version >= (60, 7, 1):
        hiddenimports += [
            'pkg_resources._vendor.jaraco.functools',
            'pkg_resources._vendor.jaraco.context',
            'pkg_resources._vendor.jaraco.text',
        ]

    # As of setuptools 70.0.0, we need pkg_resources.extern added to hidden imports.
    if setuptools_info.version >= (70, 0, 0):
        hiddenimports += [
            'pkg_resources.extern',
        ]

# Some more hidden imports. See:
# https://github.com/pyinstaller/pyinstaller-hooks-contrib/issues/15#issuecomment-663699288 `packaging` can either be
# its own package, or embedded in `pkg_resources._vendor.packaging`, or both.
hiddenimports += collect_submodules('packaging')



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Task
from waflib.Configure import conf
from waflib.TaskGen import feature, before_method, after_method

LIB_CODE = '''
#ifdef _MSC_VER
#define testEXPORT __declspec(dllexport)
#else
#define testEXPORT
#endif
testEXPORT int lib_func(void) { return 9; }
'''
MAIN_CODE = '''
#ifdef _MSC_VER
#define testEXPORT __declspec(dllimport)
#else
#define testEXPORT
#endif
testEXPORT int lib_func(void);
int main(int argc, char **argv) {
	(void)argc; (void)argv;
	return !(lib_func() == 9);
}
'''


@feature('link_lib_test')
@before_method('process_source')
def link_lib_test_fun(self):
    def write_test_file(task):
        task.outputs[0].write(task.generator.code)

    rpath = []
    if getattr(self, 'add_rpath', False):
        rpath = [self.bld.path.get_bld().abspath()]
    mode = self.mode
    m = '%s %s' % (mode, mode)
    ex = self.test_exec and 'test_exec' or ''
    bld = self.bld
    bld(rule=write_test_file, target='test.' + mode, code=LIB_CODE)
    bld(rule=write_test_file, target='main.' + mode, code=MAIN_CODE)
    bld(features='%sshlib' % m, source='test.' + mode, target='test')
    bld(features='%sprogram %s' % (m, ex), source='main.' + mode, target='app', use='test', rpath=rpath)


@conf
def check_library(self, mode=None, test_exec=True):
    if not mode:
        mode = 'c'
        if self.env.CXX:
            mode = 'cxx'
    self.check(
        compile_filename=[], features='link_lib_test', msg='Checking for libraries', mode=mode, test_exec=test_exec
    )


INLINE_CODE = '''
typedef int foo_t;
static %s foo_t static_foo () {return 0; }
%s foo_t foo () {
	return 0;
}
'''
INLINE_VALUES = ['inline', '__inline__', '__inline']


@conf
def check_inline(self, **kw):
    self.start_msg('Checking for inline')
    if not 'define_name' in kw:
        kw['define_name'] = 'INLINE_MACRO'
    if not 'features' in kw:
        if self.env.CXX:
            kw['features'] = ['cxx']
        else:
            kw['features'] = ['c']
    for x in INLINE_VALUES:
        kw['fragment'] = INLINE_CODE % (x, x)
        try:
            self.check(**kw)
        except self.errors.ConfigurationError:
            continue
        else:
            self.end_msg(x)
            if x != 'inline':
                self.define('inline', x, quote=False)
            return x
    self.fatal('could not use inline functions')


LARGE_FRAGMENT = '''#include <unistd.h>
int main(int argc, char **argv) {
	(void)argc; (void)argv;
	return !(sizeof(off_t) >= 8);
}
'''


@conf
def check_large_file(self, **kw):
    if not 'define_name' in kw:
        kw['define_name'] = 'HAVE_LARGEFILE'
    if not 'execute' in kw:
        kw['execute'] = True
    if not 'features' in kw:
        if self.env.CXX:
            kw['features'] = ['cxx', 'cxxprogram']
        else:
            kw['features'] = ['c', 'cprogram']
    kw['fragment'] = LARGE_FRAGMENT
    kw['msg'] = 'Checking for large file support'
    ret = True
    try:
        if self.env.DEST_BINFMT != 'pe':
            ret = self.check(**kw)
    except self.errors.ConfigurationError:
        pass
    else:
        if ret:
            return True
    kw['msg'] = 'Checking for -D_FILE_OFFSET_BITS=64'
    kw['defines'] = ['_FILE_OFFSET_BITS=64']
    try:
        ret = self.check(**kw)
    except self.errors.ConfigurationError:
        pass
    else:
        self.define('_FILE_OFFSET_BITS', 64)
        return ret
    self.fatal('There is no support for large files')


ENDIAN_FRAGMENT = '''
#ifdef _MSC_VER
#define testshlib_EXPORT __declspec(dllexport)
#else
#define testshlib_EXPORT
#endif

short int ascii_mm[] = { 0x4249, 0x4765, 0x6E44, 0x6961, 0x6E53, 0x7953, 0 };
short int ascii_ii[] = { 0x694C, 0x5454, 0x656C, 0x6E45, 0x6944, 0x6E61, 0 };
int testshlib_EXPORT use_ascii (int i) {
	return ascii_mm[i] + ascii_ii[i];
}
short int ebcdic_ii[] = { 0x89D3, 0xE3E3, 0x8593, 0x95C5, 0x89C4, 0x9581, 0 };
short int ebcdic_mm[] = { 0xC2C9, 0xC785, 0x95C4, 0x8981, 0x95E2, 0xA8E2, 0 };
int use_ebcdic (int i) {
	return ebcdic_mm[i] + ebcdic_ii[i];
}
extern int foo;
'''


class grep_for_endianness(Task.Task):
    color = 'PINK'

    def run(self):
        txt = self.inputs[0].read(flags='rb').decode('latin-1')
        if txt.find('LiTTleEnDian') > -1:
            self.generator.tmp.append('little')
        elif txt.find('BIGenDianSyS') > -1:
            self.generator.tmp.append('big')
        else:
            return -1


@feature('grep_for_endianness')
@after_method('apply_link')
def grep_for_endianness_fun(self):
    self.create_task('grep_for_endianness', self.link_task.outputs[0])


@conf
def check_endianness(self):
    tmp = []

    def check_msg(self):
        return tmp[0]

    self.check(
        fragment=ENDIAN_FRAGMENT,
        features='c cshlib grep_for_endianness',
        msg='Checking for endianness',
        define='ENDIANNESS',
        tmp=tmp,
        okmsg=check_msg,
        confcache=None
    )
    return tmp[0]



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import sys
from waflib.Tools import ccroot, ar, gcc
from waflib.Configure import conf


@conf
def find_icc(conf):
    cc = conf.find_program(['icc', 'ICL'], var='CC')
    conf.get_cc_version(cc, icc=True)
    conf.env.CC_NAME = 'icc'


def configure(conf):
    conf.find_icc()
    conf.find_ar()
    conf.gcc_common_flags()
    conf.gcc_modifier_platform()
    conf.cc_load_tools()
    conf.cc_add_flags()
    conf.link_add_flags()



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# NOTE: relative imports do not work beyond toplevel package.

# Should look for module 'pyi_testmod_relimp3c' in current directory.
from .pyi_testmod_relimp3c import c1
# Should look for module 'b1' in directory '../pyi_testmod_relimp3b' - one level up.
from ..pyi_testmod_relimp3b import b1


def getString():
    return b1.string + c1.string



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import TaskGen, Task
from waflib.Tools import c_preproc
from waflib.Tools.ccroot import link_task, stlink_task


@TaskGen.extension('.c')
def c_hook(self, node):
    if not self.env.CC and self.env.CXX:
        return self.create_compiled_task('cxx', node)
    return self.create_compiled_task('c', node)


class c(Task.Task):
    run_str = '${CC} ${ARCH_ST:ARCH} ${CFLAGS} ${FRAMEWORKPATH_ST:FRAMEWORKPATH} ${CPPPATH_ST:INCPATHS} ${DEFINES_ST:DEFINES} ${CC_SRC_F}${SRC} ${CC_TGT_F}${TGT[0].abspath()} ${CPPFLAGS}'
    vars = ['CCDEPS']
    ext_in = ['.h']
    scan = c_preproc.scan


class cprogram(link_task):
    run_str = '${LINK_CC} ${LINKFLAGS} ${CCLNK_SRC_F}${SRC} ${CCLNK_TGT_F}${TGT[0].abspath()} ${RPATH_ST:RPATH} ${FRAMEWORKPATH_ST:FRAMEWORKPATH} ${FRAMEWORK_ST:FRAMEWORK} ${ARCH_ST:ARCH} ${STLIB_MARKER} ${STLIBPATH_ST:STLIBPATH} ${STLIB_ST:STLIB} ${SHLIB_MARKER} ${LIBPATH_ST:LIBPATH} ${LIB_ST:LIB} ${LDFLAGS}'
    ext_out = ['.bin']
    vars = ['LINKDEPS']
    inst_to = '${BINDIR}'


class cshlib(cprogram):
    inst_to = '${LIBDIR}'


class cstlib(stlink_task):
    pass



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pytest
import textwrap

from PyInstaller.depend import utils
from PyInstaller import compat

CTYPES_CLASSNAMES = (
    'CDLL', 'ctypes.CDLL',
    'WinDLL', 'ctypes.WinDLL',
    'OleDLL', 'ctypes.OleDLL',
    'PyDLL', 'ctypes.PyDLL',
)  # yapf: disable


def __scan_code_for_ctypes(code, monkeypatch, extended_args):
    # _resolveCtypesImports would filter our some of our names
    monkeypatch.setattr(utils, '_resolveCtypesImports', lambda cbinaries: cbinaries)
    code = textwrap.dedent(code)

    if extended_args:
        # Chuck in a load of preceding rubbish to test if the bytecode scanner can correctly
        # handle the EXTENDED_ARGS opcode.
        from test_bytecode import many_str_constants, many_globals
        code = many_str_constants() + many_globals() + code

    co = compile(code, 'dummy', 'exec')
    #import pdb ; pdb.set_trace()
    return utils.scan_code_for_ctypes(co)


@pytest.mark.parametrize('classname', CTYPES_CLASSNAMES)
@pytest.mark.parametrize('extended_args', [False, True])
def test_ctypes_CDLL_call(monkeypatch, classname, extended_args):
    code = f"{classname}('somelib.xxx')"
    res = __scan_code_for_ctypes(code, monkeypatch, extended_args)
    assert res == set(['somelib.xxx'])


@pytest.mark.parametrize('classname', CTYPES_CLASSNAMES)
@pytest.mark.parametrize('extended_args', [False, True])
def test_ctypes_LibraryLoader(monkeypatch, classname, extended_args):
    # This type of usage is only valid on Windows and the lib-name will always get `.dll` appended.
    code = f"{classname.lower()}.somelib"
    res = __scan_code_for_ctypes(code, monkeypatch, extended_args)
    assert res == set(['somelib.dll'])


@pytest.mark.parametrize('classname', CTYPES_CLASSNAMES)
@pytest.mark.parametrize('extended_args', [False, True])
def test_ctypes_LibraryLoader_LoadLibrary(monkeypatch, classname, extended_args):
    code = f"{classname.lower()}.LoadLibrary('somelib.xxx')"
    res = __scan_code_for_ctypes(code, monkeypatch, extended_args)
    assert res == set(['somelib.xxx'])


@pytest.mark.parametrize('extended_args', [False, True])
@pytest.mark.skipif(compat.is_musl, reason="find_library() doesn't work on musl")
@pytest.mark.skipif(
    compat.is_macos_11 and not (compat.is_macos_11_native and compat.is_py39),
    reason="find_library() requires python >= 3.9 built with Big Sur support.",
)
def test_ctypes_util_find_library(monkeypatch, extended_args):
    # for lind_library() we need a lib actually existing on the system
    if compat.is_win:
        libname = "KERNEL32"
    else:
        libname = "c"
    code = f"ctypes.util.find_library('{libname}')"
    res = __scan_code_for_ctypes(code, monkeypatch, extended_args)
    assert res


def test_ctypes_util_find_library_as_default_argument():
    # Test-case for fix:
    # commit 55b542f135340c612a861cfcce0f86c4e5a968df
    # Author: Hartmut Goebel <h.goebel@crazy-compilers.com>
    # Date:   Thu Nov 19 14:45:30 2015 +0100
    code = """
    def locate_library(loader=ctypes.util.find_library):
        pass
    """
    code = textwrap.dedent(code)
    co = compile(code, '<ctypes_util_find_library_as_default_argument>', 'exec')
    utils.scan_code_for_ctypes(co)


@pytest.mark.linux
@pytest.mark.skipif(compat.is_termux, reason="Termux does not have libc.so")
def test_ldconfig_cache():
    utils.load_ldconfig_cache()

    if compat.is_musl:
        # load_ldconfig_cache() should be a no-op on musl because musl does not use ldconfig.
        assert not utils.LDCONFIG_CACHE
        return

    libpath = None
    for soname in utils.LDCONFIG_CACHE:
        if soname.startswith('libc.so.'):
            libpath = utils.LDCONFIG_CACHE[soname]
            break
    assert libpath, 'libc.so not found'
    assert os.path.isfile(libpath)



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------

import glob
import os
import sysconfig

from PyInstaller.compat import is_win, is_linux
from PyInstaller.utils.hooks import (
    get_module_file_attribute,
    check_requirement,
    collect_delvewheel_libs_directory,
    collect_submodules,
)

binaries = []
datas = []

# Package the DLL bundle that official scipy wheels for Windows ship The DLL bundle will either be in extra-dll on
# windows proper and in .libs if installed on a virtualenv created from MinGW (Git-Bash for example)
if is_win:
    extra_dll_locations = ['extra-dll', '.libs']
    for location in extra_dll_locations:
        dll_glob = os.path.join(os.path.dirname(get_module_file_attribute('scipy')), location, "*.dll")
        if glob.glob(dll_glob):
            binaries.append((dll_glob, "."))

# Handle delvewheel-enabled win32 wheels, which have external scipy.libs directory (scipy >= 0.9.2)
if check_requirement("scipy >= 1.9.2") and is_win:
    datas, binaries = collect_delvewheel_libs_directory('scipy', datas=datas, binaries=binaries)

# collect library-wide utility extension modules
hiddenimports = ['scipy._lib.%s' % m for m in ['messagestream', "_ccallback_c", "_fpumode"]]

# In scipy 1.14.0, `scipy._lib.array_api_compat.numpy` added a programmatic import of its `.fft` submodule, which needs
# to be added to hiddenimports.
if check_requirement("scipy >= 1.14.0"):
    hiddenimports += ['scipy._lib.array_api_compat.numpy.fft']

# If scipy is provided by Debian's python3-scipy, its scipy.__config__ submodule is renamed to a dynamically imported
# scipy.__config__${SOABI}__
# https://salsa.debian.org/python-team/packages/scipy/-/blob/1255922cf7c52b05aa44fb733449953cd9adb815/debian/patches/scipy_config_SOABI.patch
if is_linux and "dist-packages" in get_module_file_attribute("scipy"):
    hiddenimports.append('scipy.__config__' + sysconfig.get_config_var('SOABI') + '__')

# The `scipy._lib.array_api_compat.numpy` module performs a `from numpy import *`; in numpy 2.0.0, `numpy.f2py` was
# added to `numpy.__all__` attribute, but at the same time, the upstream numpy hook adds `numpy.f2py` to
# `excludedimports`. Therefore, the `numpy.f2py` sub-package ends up missing. Due to the way exclusion mechanism works,
# we need to add both `numpy.f2py` and all its submodules to hiddenimports here.
if check_requirement("numpy >= 2.0.0"):
    hiddenimports += collect_submodules('numpy.f2py', filter=lambda name: name != 'numpy.f2py.tests')

# Starting with scipy 1.17.0, the `scipy._cyutility` extension is imported when top-level `scipy` package is imported
# (via import of another extension, `scipy._lib`).
if check_requirement("scipy >= 1.17.0"):
    hiddenimports += ['scipy._cyutility']



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os

import pytest

from PyInstaller import isolated
from PyInstaller.compat import is_win, is_darwin, is_linux
from PyInstaller.utils.hooks import check_requirement, can_import_module
from PyInstaller.utils.hooks.qt import get_qt_library_info
from PyInstaller.utils.tests import importorskip, requires, skipif, onedir_only


def qt_param(qt_flavor, *args, **kwargs):
    """
    A Qt flavour to be used in @pytest.mark.parametrize(). Implicitly skips the test if said flavor is not installed.
    """
    p = pytest.param(qt_flavor, *args, **kwargs)
    return pytest.param(*p.values, marks=(requires(qt_flavor),) + p.marks, id=p.id)


# Parametrize test to run the same basic code on both Python Qt libraries.
_QT_PY_PACKAGES = ['PyQt5', 'PyQt6', 'PySide2', 'PySide6']
QtPyLibs = pytest.mark.parametrize('QtPyLib', [qt_param(i) for i in _QT_PY_PACKAGES])

# macOS .app bundles, produced by the ``--windowed`` flag, invoke a unique code path that sometimes causes failures in
# Qt applications. So build with ``--windowed`` option, which will build and run both POSIX build and .app bundle.
USE_WINDOWED_KWARG = dict(pyi_args=['--windowed']) if is_darwin else {}


# We need to ensure that all QtLibraryInfo structures in Qt hook utils are initialized at this point, before the actual
# tests start. This is to prevent test-order-dependent behavior and potential issues, and applies to all platforms.
#
# Some tests (e.g., test_import::test_import_pyqt5_uic_port) may modify search path to fake PyQt5 module, and if that
# test is the point of initialization for the corresponding QtLibraryInfo structure (triggered by hooks' access to
# .version attribute), the structure ends up with invalid data for subsequent tests as well.
#
# Former solution to this problem was clearing QtLibraryInfo.version at the end of pyi_builder() fixture, which triggers
# re-initialization in each test. But as the content of QtLibraryInfo should be immutable (save for the test with fake
# module), it seems better to pre-initialize the structures in order to ensure predictable behavior.
def _ensure_qt_library_info_is_initialized():
    for pkg in _QT_PY_PACKAGES:
        try:
            info = get_qt_library_info(pkg)
            assert info.version  # trigger initialization
        except Exception:
            pass


_ensure_qt_library_info_is_initialized()


@QtPyLibs
def test_Qt_QtWidgets(pyi_builder, QtPyLib):
    pyi_builder.test_source(
        f"""
        import sys

        from {QtPyLib}.QtWidgets import QApplication, QWidget
        from {QtPyLib}.QtCore import QTimer

        is_qt6 = '{QtPyLib}' in {{'PySide6', 'PyQt6'}}

        app = QApplication(sys.argv)
        window = QWidget()
        window.setWindowTitle('Hello world!')
        window.show()

        # Exit Qt when the main loop becomes idle.
        QTimer.singleShot(0, app.exit)

        if is_qt6:
            # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
            res = app.exec()
        else:
            res = app.exec_()
        sys.exit(res)
        """, **USE_WINDOWED_KWARG
    )


@QtPyLibs
def test_Qt_QtQml(pyi_builder, QtPyLib):
    # Qt6 6.6.3 split Qt Quick Controls 2 styles into separate shared libraries, and both PySide6 6.6.3 and PyQt6 6.6.3
    # PyPI wheels failed to account for that. Skip this test if running with affected version.
    if QtPyLib == 'PyQt6':
        # With PyQt6, the shared libraries are missing on Windows and Linux, but not on macOS.
        if not is_darwin and check_requirement('PyQt6-Qt6 == 6.6.3'):
            pytest.skip('PyQt6-Qt6 6.6.3 is missing shared libraries required by Qt Quick Controls 2.')
    if QtPyLib == 'PySide6':
        # With PySide6, all OSes seem to be affected.
        if check_requirement('PySide6-Essentials == 6.6.3'):
            pytest.skip('PySide6-Essentials 6.6.3 is missing shared libraries required by Qt Quick Controls 2.')

    pyi_builder.test_source(
        f"""
        import sys

        from {QtPyLib}.QtGui import QGuiApplication
        from {QtPyLib}.QtQml import QQmlApplicationEngine
        from {QtPyLib}.QtCore import QTimer, QUrl

        is_qt6 = '{QtPyLib}' in {{'PyQt6', 'PySide6'}}

        # Select a style via the `command line
        # <https://doc.qt.io/qt-5/qtquickcontrols2-styles.html#command-line-argument>`_,
        # as PyQt5 currently does not `support https://riverbankcomputing.com/pipermail/pyqt/2018-March/040180.html>`_
        # ``QQuickStyle``. Using this style with the QML below helps to verify that all QML files are packaged; see
        # https://github.com/pyinstaller/pyinstaller/issues/3711.
        #
        # In Qt5, the style name is lower case ('imagine'), whereas in Qt6, it is capitalized ('Imagine')
        app = QGuiApplication(sys.argv + ['-style', 'Imagine' if is_qt6 else 'imagine'])
        engine = QQmlApplicationEngine()
        engine.loadData(b'''
            import QtQuick 2.11
            import QtQuick.Controls 2.4

            ApplicationWindow {{
                visible: true
                ProgressBar {{value: 0.6}}
            }}
            ''', QUrl())

        if not engine.rootObjects():
            sys.exit(-1)

        # Exit Qt when the main loop becomes idle.
        QTimer.singleShot(0, app.exit)

        if is_qt6:
            # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
            res = app.exec()
        else:
            res = app.exec_()
        del engine
        sys.exit(res)
        """, **USE_WINDOWED_KWARG
    )


@QtPyLibs
def test_Qt_QtNetwork_SSL_support(pyi_builder, QtPyLib):
    # Skip the test if QtNetwork does not support SSL (e.g., due to lack of compatible OpenSSL shared library on the
    # test system). Starting with Qt 6.1, different backends provide TLS functionality, so explicitly check if
    # 'openssl' backend is available.
    @isolated.decorate
    def check_openssl_support(package):
        import sys
        import importlib

        QtCore = importlib.import_module('.QtCore', package)
        QtNetwork = importlib.import_module('.QtNetwork', package)

        # We must initialize QCoreApplication before using QtNetwork
        app = QtCore.QCoreApplication(sys.argv)  # noqa: F841

        if not QtNetwork.QSslSocket.supportsSsl():
            return False

        # For Qt >= 6.1, check if `openssl` TLS backend is available
        try:
            qt_version = QtCore.QLibraryInfo.version().segments()
        except AttributeError:
            qt_version = []  # Qt <= 5.8

        if qt_version < [6, 1]:
            return True  # TLS backends not implemented yet

        return 'openssl' in QtNetwork.QSslSocket.availableBackends()

    if not check_openssl_support(QtPyLib):
        pytest.skip('QtNetwork does not use OpenSSL.')

    pyi_builder.test_source(
        f"""
        import sys
        from {QtPyLib}.QtCore import QCoreApplication, QLibraryInfo
        from {QtPyLib}.QtNetwork import QSslSocket

        app = QCoreApplication(sys.argv)

        # Make sure SSL is supported
        assert QSslSocket.supportsSsl(), "SSL not supported!"

        # Display OpenSSL info
        print(
            f"OpenSSL build version: {{QSslSocket.sslLibraryBuildVersionNumber():X}} "
            f"({{QSslSocket.sslLibraryBuildVersionString()}})"
        )
        print(
            f"OpenSSL run-time version: {{QSslSocket.sslLibraryVersionNumber():X}} "
            f"({{QSslSocket.sslLibraryVersionString()}})"
        )

        # Obtain Qt version
        try:
            qt_version = QLibraryInfo.version().segments()
        except AttributeError:
            qt_version = []  # Qt <= 5.8

        # If Qt supports TLS backends (>= 6.1), make sure OpenSSL backend is available.
        if qt_version >= [6, 1]:
            print(f"Active TLS backend: {{QSslSocket.activeBackend()}}")
            print(f"Available TLS backends: {{QSslSocket.availableBackends()}}")
            assert 'openssl' in QSslSocket.availableBackends(), "OpenSSL TLS backend not available!"
        """, **USE_WINDOWED_KWARG
    )


@QtPyLibs
def test_Qt_QTranslate(pyi_builder, QtPyLib):
    pyi_builder.test_source(
        f"""
        import sys
        from {QtPyLib}.QtWidgets import QApplication
        from {QtPyLib}.QtCore import QTranslator, QLocale, QLibraryInfo

        # Initialize Qt default translations
        app = QApplication(sys.argv)
        translator = QTranslator()
        locale = QLocale('de_DE')
        if hasattr(QLibraryInfo, 'path'):
            # Qt6
            translation_path = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
        else:
            # Qt5
            translation_path = QLibraryInfo.location(QLibraryInfo.TranslationsPath)

        print('Qt locale path: %s' % translation_path)

        if translator.load(locale, "qtbase_", directory=translation_path):
            print('Qt locale %s loaded.' % locale.name())
        else:
            print('Qt locale %s not found!' % locale.name())
            assert False
        """
    )


@QtPyLibs
def test_Qt_Ui_file(pyi_builder, data_dir, QtPyLib):
    pyi_builder.test_source(
        f"""
        import os
        import sys

        import {QtPyLib}.QtQuickWidgets  # Used instead of hiddenimports

        from {QtPyLib}.QtWidgets import QApplication, QWidget
        from {QtPyLib}.QtCore import QTimer

        is_qt6 = '{QtPyLib}' in {{'PyQt6', 'PySide6'}}
        is_pyqt = '{QtPyLib}' in {{'PyQt5', 'PyQt6'}}

        app = QApplication(sys.argv)

        # In Qt6, QtQuick supports multiple render APIs and automatically selects one.
        # However, QtQuickWidgets.QQuickWidget that is used by the test UI file supports only OpenGL,
        # so we need to explicitly select it via QQuickWindow.setGraphicsApi() call.
        if is_qt6:
            try:
                # This seems to be unsupported on macOS version of PySide6 at the time of writing (6.1.0)
                from {QtPyLib}.QtQuick import QQuickWindow, QSGRendererInterface
                QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.OpenGL)
            except Exception:
                pass

        # Load the UI
        ui_file = os.path.join(os.path.dirname(__file__), 'gui.ui')
        if is_pyqt:
            # Use PyQt.uic
            from {QtPyLib} import uic
            window = QWidget()
            uic.loadUi(ui_file, window)
        else:
            # Use PySide.QtUiTools.QUiLoader
            from {QtPyLib}.QtUiTools import QUiLoader
            loader = QUiLoader()
            window = loader.load(ui_file)
        window.show()

        # Exit Qt when the main loop becomes idle.
        QTimer.singleShot(0, app.exit)

        # Run the main loop
        if is_qt6:
            # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
            res = app.exec()
        else:
            res = app.exec_()
        sys.exit(res)
        """,
        # Collect the .ui file into top-level application directory.
        pyi_args=['--add-data', f"{data_dir / 'gui.ui'}:."],
    )


# Test that the ``PyQt5.Qt`` module works by importing something from it.
#
# NOTE: the ``PyQt5.Qt`` consolidating module is specific to PyQt5. It is not present in either PySide2 nor PySide6,
# and its consolidating behavior has been removed in PyQt6.
#
# The Qt Bluetooth API (which any import to ``PyQt5.Qt`` implicitly imports) is not compatible with Windows Server
# 2012 R2, the OS Appveyor runs. Specifically, running on Server 2012 causes the test to display an error in
# `a dialog box <https://github.com/mindfulness-at-the-computer/mindfulness-at-the-computer/issues/234>`_.
# The alternative of using a newer Appveyor OS `fails <https://github.com/pyinstaller/pyinstaller/pull/3563>`_.
# Therefore, skip this test on Appveyor by testing for one of its `environment variables
# <https://www.appveyor.com/docs/environment-variables/>`_.
@skipif(os.environ.get('APPVEYOR') == 'True', reason='The Appveyor OS is incompatible with PyQt.Qt.')
@requires('PyQt5')
@pytest.mark.skipif(
    check_requirement('PyQt5 == 5.11.3') and is_darwin,
    reason='This version of the macOS wheel does not include QWebEngine.'
)
def test_PyQt5_Qt(pyi_builder):
    pyi_builder.test_source('from PyQt5.Qt import QLibraryInfo', **USE_WINDOWED_KWARG)


# QtWebEngine tests


# On linux systems with glibc >= 2.34, QtWebEngine helper process crashes with SIGSEGV due to use of `clone3` syscall,
# which is incompatible with chromium sandbox (see QTBUG-96214). The issue was fixed in Qt5 5.15.7, however even the
# latest PyPI wheels of PySide2 (5.15.2.1) and PyQt5/PyQtWebEngine (5.15.6) still seem to ship Qt5 5.15.2 (which was
# probably last publicly available linux build from the Qt itself). If we encounter incompatible combination of
# glibc and Qt5 (for example, using PyPI wheels under Ubuntu 22.04), we disable the sandbox, which allows us to perform
# basic functionality test.
def _disable_qtwebengine_sandbox(qt_flavor):
    if is_linux:
        import platform

        # Check glibc version
        libc_name, libc_version = platform.libc_ver()
        if libc_name != 'glibc':
            return False
        try:
            libc_version = [int(v) for v in libc_version.split('.')]
        except Exception:
            return False
        if libc_version < [2, 34]:
            return False

        # Check Qt version
        qt_info = get_qt_library_info(qt_flavor)
        if qt_info.version and qt_info.version >= [5, 15, 7]:
            return False

        # Incompatible glibc and Qt5 version
        return True

    return False


# Run the the QtWebEngineWidgets test for chosen Qt-based package flavor.
def _test_Qt_QtWebEngineWidgets(pyi_builder, qt_flavor):
    disable_sandbox = _disable_qtwebengine_sandbox(qt_flavor)
    pyi_builder.test_source(
        f"""
        import sys

        # Disable QtWebEngine/chromium sanbox, if necessary
        if {disable_sandbox}:
            import os
            os.environ['QTWEBENGINE_DISABLE_SANDBOX'] = '1'

        from {qt_flavor}.QtWidgets import QApplication
        from {qt_flavor}.QtWebEngineWidgets import QWebEngineView
        from {qt_flavor}.QtCore import QTimer

        is_qt6 = '{qt_flavor}' in {{'PyQt6', 'PySide6'}}

        # Web page to display
        WEB_PAGE_HTML = '''
            <!doctype html>
            <html lang="en">
                <head>
                    <meta charset="utf-8">
                    <title>Test web page</title>
                </head>
                <body>
                    <p>This is a test web page with internationalised characters.</p>
                    <p>HЯ⾀ÄÉÖÜ</p>
                </body>
            </html>
        '''

        app = QApplication(sys.argv)

        class JSResultTester:

            CODE = "1+1"
            EXPECTED = 2

            def __init__(self):
                self.result = None

            def setup(self, view):
                # Need to explicitly pass 0 as world id due to
                # https://bugreports.qt.io/browse/PYSIDE-643
                view.page().runJavaScript(self.CODE, 0, self.store)

                view.page().loadFinished.connect(
                    # Display the web page for one second after it loads.
                    lambda ok: QTimer.singleShot(1000, self.verify_and_quit))

            def store(self, res):
                self.result = res

            def verify_and_quit(self):
                # Make sure the renderer process is alive.
                print("Checking the result of renderer process...", file=sys.stderr)
                if self.result != self.EXPECTED:
                    raise ValueError(
                        f"JS result is {{self.result!r}} but expected {{self.EXPECTED!r}}. "
                        "Is the QtWebEngine renderer process running properly?")
                print("Exiting application's main loop...", file=sys.stderr)
                app.quit()

        view = QWebEngineView()
        view.setHtml(WEB_PAGE_HTML)
        view.show()

        js_result_tester = JSResultTester()
        js_result_tester.setup(view)

        print("Entering application's main loop...", file=sys.stderr)
        if is_qt6:
            # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
            res = app.exec()
        else:
            res = app.exec_()
        print("Exited application's main loop!", file=sys.stderr)

        print("Calling sys.exit()...", file=sys.stderr)
        sys.exit(res)
        """, **USE_WINDOWED_KWARG
    )


# Run the the QtWebEngineQuick test for chosen Qt-based package flavor.
def _test_Qt_QtWebEngineQuick(pyi_builder, qt_flavor):
    disable_sandbox = _disable_qtwebengine_sandbox(qt_flavor)
    pyi_builder.test_source(
        f"""
        import sys

        # Disable QtWebEngine/chromium sanbox, if necessary
        if {disable_sandbox}:
            import os
            os.environ['QTWEBENGINE_DISABLE_SANDBOX'] = '1'

        from {qt_flavor}.QtGui import QGuiApplication
        from {qt_flavor}.QtQml import QQmlApplicationEngine

        is_qt6 = '{qt_flavor}' in {{'PyQt6', 'PySide6'}}

        if is_qt6:
            from {qt_flavor}.QtWebEngineQuick import QtWebEngineQuick
        else:
            from {qt_flavor}.QtWebEngine import QtWebEngine as QtWebEngineQuick

        # Must be called before QGuiApplication is instantiated!
        QtWebEngineQuick.initialize()

        app = QGuiApplication(sys.argv)
        engine = QQmlApplicationEngine()
        engine.loadData(b'''
            import QtQuick 2.0
            import QtQuick.Window 2.0
            import QtWebEngine 1.0

            Window {{
                visible: true
                WebEngineView {{
                    id: view
                    anchors.fill: parent
                    Component.onCompleted: () => {{
                        console.info("Loading HTML...")
                        loadHtml('
                            <!doctype html>
                            <html lang="en">
                                <head>
                                    <meta charset="utf-8">
                                    <title>Test web page</title>
                                </head>
                                <body>
                                    <p>This is a test web page.</p>
                                </body>
                            </html>
                        ')
                    }}
                    onLoadingChanged: (loadRequest) => {{
                        console.info("Web page loading status changed: " + loadRequest.status)
                        if (loadRequest.status !== WebEngineView.LoadStartedStatus) {{
                            console.info("Page loading finished; running shutdown timer (" + timer.interval + " ms)!")
                            timer.running = true
                        }}
                    }}
                }}
                Timer {{
                    id: timer
                    interval: 1000
                    running: false
                    repeat: false
                    onTriggered: () => {{
                        console.info("Shutdown timer triggered; exiting application's main loop...")
                        Qt.quit()
                    }}
                }}
            }}
        ''')

        if not engine.rootObjects():
            raise RuntimeError("No root objects loaded from QML!")

        print("Entering application's main loop...", file=sys.stderr)
        if is_qt6:
            # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
            res = app.exec()
        else:
            res = app.exec_()
        print("Exited application's main loop!", file=sys.stderr)

        del engine

        print("Calling sys.exit()...", file=sys.stderr)
        sys.exit(res)
        """, **USE_WINDOWED_KWARG
    )


@requires('PyQt5')
@requires('PyQtWebEngine')
def test_Qt_QtWebEngineWidgets_PyQt5(pyi_builder):
    _test_Qt_QtWebEngineWidgets(pyi_builder, 'PyQt5')


@requires('PyQt5')
@requires('PyQtWebEngine')
def test_Qt_QtWebEngineQuick_PyQt5(pyi_builder):
    _test_Qt_QtWebEngineQuick(pyi_builder, 'PyQt5')


@requires('PySide2')
def test_Qt_QtWebEngineWidgets_PySide2(pyi_builder):
    _test_Qt_QtWebEngineWidgets(pyi_builder, 'PySide2')


@requires('PySide2')
def test_Qt_QtWebEngineQuick_PySide2(pyi_builder):
    _test_Qt_QtWebEngineQuick(pyi_builder, 'PySide2')


@requires('PyQt6 >= 6.2.2')
@requires('PyQt6-WebEngine')  # NOTE: base Qt6 must be 6.2.2 or newer, QtWebEngine can be older
@pytest.mark.flaky(
    # Attempt to mitigate issues with QtWebEngine 6.10.1
    condition=check_requirement('PyQt6-WebEngine-Qt6 == 6.10.1'),
    reruns=1,
)
def test_Qt_QtWebEngineWidgets_PyQt6(pyi_builder):
    _test_Qt_QtWebEngineWidgets(pyi_builder, 'PyQt6')


@requires('PyQt6 >= 6.2.2')
@requires('PyQt6-WebEngine')  # NOTE: base Qt6 must be 6.2.2 or newer, QtWebEngine can be older
@pytest.mark.skipif(
    check_requirement('PyQt6-Qt6 == 6.6.0'),
    reason='PyQt6 6.6.0 PyPI wheels are missing Qt6WebChannelQuick shared library.'
)
@pytest.mark.skipif(
    check_requirement('PyQt6-Qt6 == 6.6.3') and is_win,
    reason='PyQt6 6.6.3 PyPI wheels for Windows are missing Qt6WebChannelQuick shared library.'
)
@pytest.mark.flaky(
    # The generated .app bundle seems to sporadically freeze during shutdown on GHA macos-14 runners.
    # Attempt to mitigate issues with QtWebEngine 6.10.1
    condition=is_darwin or check_requirement('PyQt6-WebEngine-Qt6 == 6.10.1'),
    reruns=1,
)
def test_Qt_QtWebEngineQuick_PyQt6(pyi_builder):
    _test_Qt_QtWebEngineQuick(pyi_builder, 'PyQt6')


@requires('PySide6 >= 6.2.2')
@pytest.mark.skipif(
    check_requirement('PySide6 == 6.5.0') and is_win,
    reason='PySide6 6.5.0 PyPI wheels for Windows are missing opengl32sw.dll.'
)
@pytest.mark.flaky(
    # Attempt to mitigate issues with QtWebEngine 6.10.1
    condition=check_requirement('PySide6 == 6.10.1'),
)
def test_Qt_QtWebEngineWidgets_PySide6(pyi_builder):
    _test_Qt_QtWebEngineWidgets(pyi_builder, 'PySide6')


@requires('PySide6 >= 6.2.2')
@pytest.mark.skipif(
    check_requirement('PySide6 == 6.5.0') and is_win,
    reason='PySide6 6.5.0 PyPI wheels for Windows are missing opengl32sw.dll.'
)
@pytest.mark.flaky(
    # The generated .app bundle seems to sporadically freeze during shutdown on GHA macos-14 runners.
    # Attempt to mitigate issues with QtWebEngine 6.10.1
    condition=is_darwin or check_requirement('PySide6 == 6.10.1'),
    reruns=1,
)
def test_Qt_QtWebEngineQuick_PySide6(pyi_builder):
    _test_Qt_QtWebEngineQuick(pyi_builder, 'PySide6')


# QtMultimedia test that triggers error when the module's plugins are missing (#7352).
@QtPyLibs
def test_Qt_QtMultimedia_player_init(pyi_builder, QtPyLib):
    # QtMultimedia in PyQt6-Qt6 6.7.1 does not seem to be compatible with PyQt6 6.7.0 (the PyQt6-Qt6 6.7.1 update was
    # pushed on Jun 2 2024 without PyQt6 itself being updated).
    if QtPyLib == 'PyQt6':
        if check_requirement('PyQt6-Qt6 == 6.7.1') and check_requirement('PyQt6 == 6.7.0'):
            pytest.skip('QtMultimedia is broken under PyQt6 6.7.0 and PyQt6-Qt6 6.7.1.')

    pyi_builder.test_source(
        f"""
        import sys

        from {QtPyLib} import QtCore, QtMultimedia

        app = QtCore.QCoreApplication(sys.argv)
        player = QtMultimedia.QMediaPlayer(app)
        """, **USE_WINDOWED_KWARG
    )


# QtMultimedia test that also uses PySide's true_property, which triggers hidden dependency on QtMultimediaWidgets
# python module.
# See:
# https://github.com/pyinstaller/pyinstaller/pull/6496#issuecomment-1011098019
# https://github.com/qtproject/pyside-pyside-setup/blob/5.15.2/sources/shiboken2/shibokenmodule/files.dir/shibokensupport/signature/mapping.py#L577-L586
# https://github.com/qtproject/pyside-pyside-setup/blob/v6.2.2.1/sources/shiboken6/shibokenmodule/files.dir/shibokensupport/signature/mapping.py#L614-L627
@pytest.mark.parametrize('QtPyLib', [
    qt_param('PySide2'),
    qt_param('PySide6'),
])
def test_Qt_QtMultimedia_with_true_property(pyi_builder, QtPyLib):
    pyi_builder.test_source(
        f"""
        import sys
        from {QtPyLib} import QtCore, QtMultimedia
        from __feature__ import true_property

        app = QtCore.QCoreApplication(sys.argv)
        """, **USE_WINDOWED_KWARG
    )


# In PySide6 >= 6.4.0, we need to collect `PySide6.support.deprecated` module for logical operators between Qt key and
# key modifier enums to work. See #7249.
@requires('PySide6')
def test_Qt_PySide6_key_enums(pyi_builder):
    pyi_builder.test_source(
        """
        from PySide6 import QtCore
        key = QtCore.Qt.AltModifier | QtCore.Qt.Key_D
        """
    )


# Basic import tests for all Qt-bindings-provided modules. Each module should be importable on its own, which requires a
# corresponding hook that performs recursive analysis of the module in order to collect all of its dependencies.
#
# Due to the sheer amount of tests, they are ran only in onedir mode.


# Helper that lists all Qt* modules from a Qt-based package. Ran isolated to prevent import affecting the main process.
@isolated.decorate
def _list_all_qt_submodules(package_name):
    import importlib
    import pkgutil

    try:
        package = importlib.import_module(package_name)
    except Exception:
        return []

    return sorted([
        module_info.name for module_info in pkgutil.iter_modules(package.__path__) if module_info.name.startswith("Qt")
    ])


def _test_qt_bindings_import(bindings, module, pyi_builder_onedir):
    # Check if particular module is importable. This guards against errors if a module is unavailable in particular
    # version of the bindings, or if it is provided by an extra package that is not installed.
    modname = bindings + "." + module
    if not can_import_module(modname):
        pytest.skip(f"Module '{modname}' cannot be imported.")
    # Basic import test
    # The import of the tested module is preceeded by import of the QtCore. This seems to prevent segfaults on macOS
    # with certain modules in the frozen test (QtSensors in PySide2 and PyQt5, QtWebEngine* in PySide6, etc.). The
    # segfaults occur when trying to resolve bundle identifier, which may be related to PyInstaller failing to
    # preserve the .framework bundle structure for Qt shared libraries (and importing QtCore somehow works around
    # that). Since QtCore is practically a dependency of all other Qt modules, its import does not affect the results
    # of the test much.
    pyi_builder_onedir.test_source(f"""
        import {bindings}.QtCore
        import {modname}
        """)


@importorskip('PySide2')
@pytest.mark.parametrize('module', _list_all_qt_submodules('PySide2'))
@onedir_only
def test_qt_module_import_PySide2(module, pyi_builder):
    _test_qt_bindings_import("PySide2", module, pyi_builder)


@importorskip('PySide6')
@pytest.mark.parametrize('module', _list_all_qt_submodules('PySide6'))
@onedir_only
def test_qt_module_import_PySide6(module, pyi_builder):
    _test_qt_bindings_import("PySide6", module, pyi_builder)


@importorskip('PyQt5')
@pytest.mark.parametrize('module', _list_all_qt_submodules('PyQt5'))
@onedir_only
def test_qt_module_import_PyQt5(module, pyi_builder):
    _test_qt_bindings_import("PyQt5", module, pyi_builder)


@importorskip('PyQt6')
@pytest.mark.parametrize('module', _list_all_qt_submodules('PyQt6'))
@onedir_only
def test_qt_module_import_PyQt6(module, pyi_builder):
    _test_qt_bindings_import("PyQt6", module, pyi_builder)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This contains tests for the class:``TOC``, see
# https://pyinstaller.readthedocs.io/en/latest/advanced-topics.html#the-toc-and-tree-classes

import pytest

from PyInstaller.building.datastruct import TOC

ELEMS1 = (
    ('encodings', '/usr/lib/python2.7/encodings/__init__.py', 'PYMODULE'),
    ('_random', '/usr/lib/python2.7/lib-dynload/_random.so', 'EXTENSION'),
    ('libreadline.so.6', '/lib64/libreadline.so.6', 'BINARY'),
)

ELEMS2 = (
    ('li-la-lu', '/home/myself/li-la-su', 'SOMETHING'),
    ('schubidu', '/home/otherguy/schibidu', 'PKG'),
)

ELEMS3 = (('PIL.Image.py', '/usr/lib/python2.7/encodings/__init__.py', 'PYMODULE'),)

# Ignore deprecation warnings for the TOC class
pytestmark = pytest.mark.filterwarnings("ignore:TOC class is deprecated.")


def test_init_empty():
    toc = TOC()
    assert len(toc) == 0


def test_init():
    toc = TOC(ELEMS1)
    assert len(toc) == 3
    assert toc == list(ELEMS1)


def test_append():
    toc = TOC(ELEMS1)
    toc.append(('li-la-lu', '/home/myself/li-la-su', 'SOMETHING'))
    expected = list(ELEMS1)
    expected.append(('li-la-lu', '/home/myself/li-la-su', 'SOMETHING'))
    assert toc == expected


def test_append_existing():
    toc = TOC(ELEMS1)
    toc.append(ELEMS1[-1])
    expected = list(ELEMS1)
    assert toc == expected


def test_append_keep_filename():
    # name in TOC should be the same as the one added
    toc = TOC()
    entry = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'BINARY')
    toc.append(entry)
    assert toc[0][0] == entry[0]


def test_insert():
    toc = TOC(ELEMS1)
    toc.insert(1, ('li-la-lu', '/home/myself/li-la-su', 'SOMETHING'))
    expected = list(ELEMS1)
    expected.insert(1, ('li-la-lu', '/home/myself/li-la-su', 'SOMETHING'))
    assert toc == expected


def test_insert_existing():
    toc = TOC(ELEMS1)
    toc.insert(0, ELEMS1[-1])
    toc.insert(1, ELEMS1[-1])
    expected = list(ELEMS1)
    assert toc == expected


def test_insert_keep_filename():
    # name in TOC should be the same as the one added
    toc = TOC()
    entry = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'BINARY')
    toc.insert(1, entry)
    assert toc[0][0] == entry[0]


def test_extend():
    toc = TOC(ELEMS1)
    toc.extend(ELEMS2)
    expected = list(ELEMS1)
    expected.extend(ELEMS2)
    assert toc == expected


def test_extend_existing():
    toc = TOC(ELEMS1)
    toc.extend(ELEMS1)
    expected = list(ELEMS1)
    assert toc == expected


def test_add_list():
    toc = TOC(ELEMS1)
    other = list(ELEMS2)
    result = toc + other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1) + list(ELEMS2)
    assert result == expected


def test_add_tuple():
    toc = TOC(ELEMS1)
    other = ELEMS2
    result = toc + other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1) + list(ELEMS2)
    assert result == expected


def test_add_toc():
    toc = TOC(ELEMS1)
    other = TOC(ELEMS2)
    result = toc + other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1) + list(ELEMS2)
    assert result == expected


def test_radd_list():
    toc = TOC(ELEMS1)
    other = list(ELEMS2)
    result = other + toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2) + list(ELEMS1)
    assert result == expected


def test_radd_tuple():
    toc = TOC(ELEMS1)
    other = ELEMS2
    result = other + toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2) + list(ELEMS1)
    assert result == expected


def test_radd_toc():
    toc = TOC(ELEMS1)
    other = TOC(ELEMS2)
    result = other + toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2) + list(ELEMS1)
    assert result == expected


def test_sub_list():
    toc = TOC(ELEMS1) + ELEMS2
    other = list(ELEMS2)
    result = toc - other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1)
    assert result == expected


def test_sub_tuple():
    toc = TOC(ELEMS1) + ELEMS2
    other = ELEMS2
    result = toc - other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1)
    assert result == expected


def test_sub_toc():
    toc = TOC(ELEMS1) + ELEMS2
    other = TOC(ELEMS2)
    result = toc - other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1)
    assert result == expected


def test_sub_non_existing():
    toc = TOC(ELEMS1)
    other = ELEMS3
    result = toc - other
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1)
    assert result == expected


def test_rsub_list():
    toc = TOC(ELEMS1)
    other = list(ELEMS1) + list(ELEMS2)
    result = other - toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2)
    assert result == expected


def test_rsub_tuple():
    toc = TOC(ELEMS1)
    other = tuple(list(ELEMS1) + list(ELEMS2))
    result = other - toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2)
    assert result == expected


def test_rsub_toc():
    toc = TOC(ELEMS1)
    other = TOC(ELEMS1) + ELEMS2
    result = other - toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS2)
    assert result == expected


def test_rsub_non_existing():
    toc = TOC(ELEMS3)
    other = ELEMS1
    result = other - toc
    assert result is not toc
    assert result is not other
    assert isinstance(result, TOC)
    expected = list(ELEMS1)
    assert result == expected


def test_sub_after_setitem():
    toc = TOC(ELEMS1)
    toc[1] = ('lib-dynload/_random', '/usr/lib/python2.7/lib-dynload/_random.so', 'EXTENSION')
    toc -= []
    assert len(toc) == 3


def test_sub_after_sub():
    toc = TOC(ELEMS1)
    toc -= [ELEMS1[0]]
    toc -= [ELEMS1[1]]
    expected = list(ELEMS1[2:])
    assert toc == expected


def test_setitem_1():
    toc = TOC()
    toc[:] = ELEMS1
    for e in ELEMS1:
        assert e in toc
        assert e[0] in toc.filenames


def test_setitem_2():
    toc = TOC(ELEMS1)
    toc[1] = ELEMS3[0]

    assert ELEMS1[0] in toc
    assert ELEMS1[0][0] in toc.filenames

    assert ELEMS3[0] in toc
    assert ELEMS3[0][0] in toc.filenames

    assert ELEMS1[2] in toc
    assert ELEMS1[2][0] in toc.filenames

    for e in toc:
        assert e[0] in toc.filenames


# The following tests verify that case-insensitive comparisons are used on Windows and only for
# appropriate TOC entry types


@pytest.mark.win32
def test_append_other_case_mixed():
    # If a binary file is added with the same filename as an existing pymodule, it should not be added.
    toc = TOC(ELEMS1)
    elem = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'BINARY')
    toc.append(elem)
    expected = list(ELEMS1)
    assert toc == expected


@pytest.mark.win32
def test_append_other_case_pymodule():
    # Python modules should not use C-I comparisons. Both 'encodings' and 'EnCodIngs' should be added.
    toc = TOC(ELEMS1)
    elem = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'PYMODULE')
    toc.append(elem)
    expected = list(ELEMS1)
    expected.append(elem)
    assert toc == expected


@pytest.mark.win32
def test_append_other_case_binary():
    # Binary files should use C-I comparisons. 'LiBrEADlInE.so.6' should not be added.
    toc = TOC(ELEMS1)
    toc.append(('LiBrEADlInE.so.6', '/lib64/libreadline.so.6', 'BINARY'))
    expected = list(ELEMS1)
    assert toc == expected


@pytest.mark.win32
def test_insert_other_case_mixed():
    # If a binary file is added with the same filename as an existing pymodule, it should not be added.
    toc = TOC(ELEMS1)
    elem = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'BINARY')
    toc.insert(1, elem)
    expected = list(ELEMS1)
    assert toc == expected


@pytest.mark.win32
def test_insert_other_case_pymodule():
    # Python modules should not use C-I comparisons. Both 'encodings' and 'EnCodIngs' should be added.
    toc = TOC(ELEMS1)
    elem = ('EnCodIngs', '/usr/lib/python2.7/encodings.py', 'PYMODULE')
    toc.insert(1, elem)
    expected = list(ELEMS1)
    expected.insert(1, elem)
    assert toc == expected


@pytest.mark.win32
def test_insert_other_case_binary():
    # Binary files should use C-I comparisons. 'LiBrEADlInE.so.6' should not be added.
    toc = TOC(ELEMS1)
    toc.insert(1, ('LiBrEADlInE.so.6', '/lib64/libreadline.so.6', 'BINARY'))
    expected = list(ELEMS1)
    assert toc == expected



import dis
import inspect


def iterate_instructions(code_object):
    """Delivers the byte-code instructions as a continuous stream.

    Yields `dis.Instruction`. After each code-block (`co_code`), `None` is
    yielded to mark the end of the block and to interrupt the steam.
    """
    # The arg extension the EXTENDED_ARG opcode represents is automatically handled by get_instructions() but the
    # instruction is left in. Get rid of it to make subsequent parsing easier/safer.
    yield from (i for i in dis.get_instructions(code_object) if i.opname != "EXTENDED_ARG")

    yield None

    # For each constant in this code object that is itself a code object,
    # parse this constant in the same manner.
    for constant in code_object.co_consts:
        if inspect.iscode(constant):
            yield from iterate_instructions(constant)



# Two modules; one is collected into PYZ archive (and is thus handled by PyInstaller's frozen importer); the other is
# collected as source .py file only (forced by module_collection_mode setting in the accompanying hook), and is thus
# handled by python's own `_frozen_importlib_external.PathFinder`. Both, however, need to be importable in the frozen
# application. In other words, PyInstaller's frozen importer and python's `_frozen_importlib_external.PathFinder` must
# complement, and not exclude, each other.
#
# NOTE: same scenario could be achieved with a package that contains a pure-python module (collected into PYZ by
# default) and a binary extension module (always collected as a file); however, test setup is more practical with two
# (empty) pure-python modules and manipulated collection mode.
from . import a  # noqa: F401
from . import b  # noqa: F401



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pathlib

import pytest

from PyInstaller.utils.tests import importorskip, requires

# Run the tests here in onedir mode only - onefile offers no additional insights in this context.
pytestmark = pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


# Test module exclusion; ensure that excluded modules are not collected. When exclusion is performed via hooks, also
# test that hooks for excluded modules are not ran (by having hooks raise exceptions).
@pytest.mark.parametrize(
    "exclude_args,exclude_hooks", (
        pytest.param(True, False, id='args'),
        pytest.param(False, True, id='hooks'),
        pytest.param(True, True, id='args-and-hooks'),
    )
)
def test_module_exclusion(exclude_args, exclude_hooks, pyi_builder):
    pyi_args = ['--paths', str(_MODULES_DIR / 'pyi_module_exclusion' / 'modules')]
    if exclude_args:
        pyi_args += ['--exclude', 'mymodule_feature2', '--exclude', 'mymodule_feature3']
    if exclude_hooks:
        pyi_args += ['--additional-hooks-dir', str(_MODULES_DIR / 'pyi_module_exclusion' / 'hooks')]

    pyi_builder.test_source(
        """
        import mymodule_main

        # Feature #1 module should be included, and thus available
        assert mymodule_main.feature1_available == True

        # Feature #2 module should be excluded, and thus unavailable
        assert mymodule_main.feature2_available == False

        # Feature #3 module should be excluded, and thus unavailable
        assert mymodule_main.feature3_available == False
        """,
        pyi_args=pyi_args,
        run_from_path=True
    )


# Tests for excluded subpackage from the top-level package hook. Ensure that such subpackage is excluded when the only
# reference comes from within the corresponding top-level package (or one of its submodules/subpackages). However, if
# the subpackage is referred from external source (e.g., user's program), then it should be collected and the exclusion
# rule from the top-level package hook should not block collection of submodules from that subpackage.
@pytest.mark.parametrize("with_reference", [False, True], ids=["noref", "ref"])
def test_subpackage_exclusion(pyi_builder, with_reference):
    pyi_args = [
        '--paths',
        str(_MODULES_DIR / 'pyi_excluded_subpackage' / 'modules'),
        '--additional-hooks-dir',
        str(_MODULES_DIR / 'pyi_excluded_subpackage' / 'hooks'),
    ]

    if with_reference:
        # Explicit reference to excluded subpackage in user's program; we expect the subpackage to be fully collected.
        source_code = (
            """
            import mypackage.optional  # is importable if its submodules are also collected.
            assert mypackage.optional.optional_function() == 42
            """
        )
    else:
        # No reference to excluded subpackage other than in top-level package; we expect the subpackage to be excluded.
        source_code = (
            """
            import importlib.util

            import mypackage

            # Ensure the optional subpackage is unavailable
            optional_spec = importlib.util.find_spec('mypackage.optional')
            assert optional_spec == None, "mypackage.optional is collected, but should not be!"
            """
        )

    pyi_builder.test_source(
        source_code,
        pyi_args=pyi_args,
    )


# Our hook for sqlalchemy excludes sqlalchemy.testing. Make sure explicitly importing the said subpackage works.
# See #9193.
@importorskip('sqlalchemy')
def test_subpackage_exclusion_sqlalchemy_testing(pyi_builder):
    pyi_builder.test_source("""
        import sqlalchemy.testing
        """)


# Our hook for numpy excludes numpy.f2py for numpy < 2.0. Make sure explicitly importing the said subpackage works.
@requires('numpy < 2.0')
def test_subpackage_exclusion_numpy_f2py(pyi_builder):
    pyi_builder.test_source("""
        import numpy.f2py
        """)





try:
    import os.path
except ImportError:
    pass

import os

def function(self):
    import sys


if a == b:
    import sys

def function2(self):
    if a == b:
        import platform

def function3(self):
    import platform
    import email


import email



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pathlib

from PyInstaller.utils.tests import importorskip
from PyInstaller.compat import is_darwin, exec_python_rc

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'

# All tests in this file require pkg_resources
pytestmark = importorskip('pkg_resources')


def test_pkg_resources_importable(pyi_builder):
    """
    Check that a trivial example using pkg_resources does build.
    """
    pyi_builder.test_source("""
        import pkg_resources
        pkg_resources.working_set.require()
        """)


def test_pkg_resources_resource_string(pyi_builder):
    add_data_arg = f"{_MODULES_DIR / 'pkg3' / 'sample-data.txt'}:pkg3"
    pyi_builder.test_source(
        """
        import pkg_resources
        import pkg3

        expected_data = b'This is data text for testing the packaging module data.'

        # In a frozen app, the resources is available at: os.path.join(sys._MEIPASS, 'pkg3/sample-data.txt')
        data = pkg_resources.resource_string(pkg3.__name__, 'sample-data.txt')
        if not data:
            raise SystemExit('Error: Could not read data with pkg_resources.resource_string().')

        if data.strip() != expected_data:
            raise SystemExit('Error: Read data does not match expected data!')
        """,
        pyi_args=['--add-data', add_data_arg],
    )


# Tests for pkg_resources provider.
#
# These tests run a test script (scripts/pyi_pkg_resources_provider.py) in unfrozen and frozen form, in combination with
# a custom test package (modules/pyi_pkg_resources_provider/package).
#
# Running the unfrozen test script allows us to verify the behavior of DefaultProvider from pkg_resources and thereby
# also validate the test script itself. Running the frozen test validates the behavior of the PyiFrozenProvider.
#
# For details on the structure of the test and the contents of the test package, see the top comment in the test script
# itself.
def _exec_python_script(script_filename, pathex):
    # Prepare the environment - default to 'os.environ'...
    env = os.environ.copy()
    # ... and prepend PYTHONPATH with pathex
    if 'PYTHONPATH' in env:
        pathex = os.pathsep.join([str(pathex), env['PYTHONPATH']])
    env['PYTHONPATH'] = str(pathex)
    # Run the test script
    return exec_python_rc(str(script_filename), env=env)


def test_pkg_resources_provider_source(tmp_path, script_dir, monkeypatch):
    # Run the test script unfrozen - to validate it is working and to verify the behavior of
    # pkg_resources.DefaultProvider.
    pathex = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'package'
    test_script = script_dir / 'pyi_pkg_resources_provider.py'
    ret = _exec_python_script(test_script, pathex=pathex)
    assert ret == 0, "Test script failed!"


def test_pkg_resources_provider_frozen(pyi_builder, tmp_path, script_dir, monkeypatch):
    # Run the test script as a frozen program
    pathex = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'package'
    test_script = 'pyi_pkg_resources_provider.py'
    hooks_dir = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'hooks'
    pyi_args = [
        '--paths', str(pathex),
        '--hidden-import', 'pyi_pkgres_testpkg',
        '--additional-hooks-dir', str(hooks_dir),
    ]  # yapf: disable
    if is_darwin:
        pyi_args += ['--windowed']  # Also build and test .app bundle executable
    pyi_builder.test_script(
        test_script,
        pyi_args=pyi_args,
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
import sys

excludedimports = []

# see https://github.com/python/cpython/blob/3.9/Lib/platform.py#L411
# This will exclude `plistlib` for sys.platform != 'darwin'
if sys.platform != 'darwin':
    excludedimports += ["plistlib"]

# Avoid collecting `_ios_support`, which in turn triggers unnecessary collection of `libobjc` shared library if the
# latter happens to be available on the build system.
# See https://github.com/pyinstaller/pyinstaller/issues/9333, as well as
# https://github.com/python/cpython/blob/v3.13.0/Lib/platform.py#L508-L521
# and
# https://github.com/python/cpython/blob/v3.13.0/Lib/_ios_support.py#L13
if sys.platform != 'ios':
    excludedimports += ["_ios_support"]
